import numpy as np
from ...const import BOARD_SIZE, I_COUNT, L_COUNT

dx, dy = [0, 0, 1, -1], [1, -1, 0, 0]


def bfs(self, stk, pl):
    visited = np.zeros((BOARD_SIZE, BOARD_SIZE), dtype=bool)
    visited[stk[0][0] - 1, stk[0][1] - 1] = True
    while len(stk) > 0:
        x, y = stk.pop()
        for i in range(4):
            if self.movable((x + dx[i], y + dy[i]), (x, y)) and not visited[x + dx[i] - 1, y + dy[i] - 1]:
                if pl == 0 and x + dx[i] == BOARD_SIZE:
                    return True
                elif pl == 1 and y + dy[i] == BOARD_SIZE:
                    return True
                visited[x - 1, y - 1] = True
                stk.append((x + dx[i], y + dy[i]))
    return False


def endable(self):
    return bfs(self, [self.position()], self.player()) and bfs(self, [self.position(True)], self.player(True))



from ...const import BOARD_SIZE, I_COUNT, L_COUNT


def cross(self, t, x, y, w):
    if t == 1:
        if w == 1:
            for i in range(2, 2 + I_COUNT * 2):
                if self.st[i, 0] == x - 1 and self.st[i, 1] == y + 1 and self.st[i, 2] == 2: return True
            return self.horizontal_block(x, y) or self.horizontal_block(x + 1, y)
        elif w == 2:
            for i in range(2, 2 + I_COUNT * 2):
                if self.st[i, 0] == x + 1 and self.st[i, 1] == y - 1 and self.st[i, 2] == 1: return True
            return self.vertical_block(x, y) or self.vertical_block(x, y + 1)
        else:
            return True
    elif t == 2:
        if w == 1:
            for i in range(2 + I_COUNT * 2, 2 + (I_COUNT + L_COUNT) * 2):
                if self.st[i, 0] == x - 1 and self.st[i, 1] == y - 1 and self.st[i, 2] == 3: return True
            return self.vertical_block(x, y) or self.horizontal_block(x, y)
        elif w == 2:
            for i in range(2 + I_COUNT * 2, 2 + (I_COUNT + L_COUNT) * 2):
                if self.st[i, 0] == x - 1 and self.st[i, 1] == y + 1 and self.st[i, 2] == 4: return True
            return self.horizontal_block(x, y) or self.vertical_block(x, y + 1)
        elif w == 3:
            for i in range(2 + I_COUNT * 2, 2 + (I_COUNT + L_COUNT) * 2):
                if self.st[i, 0] == x + 1 and self.st[i, 1] == y + 1 and self.st[i, 2] == 1: return True
            return self.vertical_block(x, y + 1) or self.horizontal_block(x + 1, y)
        elif w == 4:
            for i in range(2 + I_COUNT * 2, 2 + (I_COUNT + L_COUNT) * 2):
                if self.st[i, 0] == x + 1 and self.st[i, 1] == y - 1 and self.st[i, 2] == 2: return True
            return self.horizontal_block(x + 1, y) or self.vertical_block(x, y)
        else:
            return True
    return True



from ...const import BOARD_SIZE, I_COUNT, L_COUNT


def place_i(self, x, y, w):
    if w == 1:
        if x < 1 or x > BOARD_SIZE + 1 or y < 1 or y > BOARD_SIZE - 1: return None
    elif w == 2:
        if x < 1 or x > BOARD_SIZE - 1 or y < 1 or y > BOARD_SIZE + 1: return None
    else:
        return None
    idx = 2 + self.player() * I_COUNT
    fl = -1
    for i in range(I_COUNT):
        if self.st[idx + i, 0] == 0:
            fl = i
            break
    if fl == -1: return None
    if self.cross(1, x, y, w): return None
    if not self.internal() and self.block_internal(1, x, y, w): return None
    st = self.copy()
    st.st[idx + fl, 0] = x
    st.st[idx + fl, 1] = y
    st.st[0, 2] = 1 - st.st[0, 2]
    st.st[idx + fl, 2] = w
    if not st.endable(): return False
    return st


def place_l(self, x, y, w):
    if x < 1 or x > BOARD_SIZE or y < 1 or y > BOARD_SIZE: return None
    if w < 1 or w > 4: return None
    idx = 2 + 2 * I_COUNT + self.player() * L_COUNT
    fl = -1
    for i in range(L_COUNT):
        if self.st[idx + i, 0] == 0:
            fl = i
            break
    if fl == -1: return None
    if self.cross(2, x, y, w): return None
    if not self.internal() and self.block_internal(2, x, y, w): return None
    st = self.copy()
    st.st[idx + fl, 0] = x
    st.st[idx + fl, 1] = y
    st.st[0, 2] = 1 - st.st[0, 2]
    st.st[idx + fl, 2] = w
    if not st.endable(): return False
    return st



from ...const import BOARD_SIZE, I_COUNT, L_COUNT


def movable(self, pos, ppos=None, opos=(0, 0)):
    x, y = pos
    if not ppos:
        (px, py) = self.position()
        (ox, oy) = self.position(1)
    else:
        px, py = ppos
        ox, oy = opos
    if x < 1 or x > BOARD_SIZE or y < 1 or y > BOARD_SIZE:
        return False
    dis = abs(px - x) + abs(py - y)
    if dis > 2 or dis == 0:
        return False
    elif dis == 1:
        if x == px:
            if self.vertical_block(x, max(y, py)):
                return False
        elif self.horizontal_block(max(x, px), y):
            return False
    else:
        if x == px:
            yt = oy
            if ox != x or yt * 2 != y + py:
                return False
            if self.vertical_block(x, yt) or self.vertical_block(x, yt + 1):
                return False
        elif y == py:
            xt = ox
            if oy != y or xt * 2 != x + px:
                return False
            if self.horizontal_block(xt, y) or self.horizontal_block(xt + 1, y):
                return False
        elif px == ox and y == oy:
            if abs(x - px) != 1 or \
                    self.vertical_block(px, max(y, py)) or \
                    self.horizontal_block(max(x, px), y) or \
                    not self.vertical_block(px, y + 1 if y > py else y):
                return False
        elif py == oy and x == ox:
            if abs(y - py) != 1 or \
                    self.vertical_block(x, max(y, py)) or \
                    self.horizontal_block(max(x, px), y) or \
                    not self.horizontal_block(x + 1 if x > px else x, py):
                return False
        else:
            return False

    if ox == x and oy == y and not self.internal():
        return False

    return True



def block_internal(self, t, x, y, w):
    if t == 1:
        if w == 1:
            if 5 <= x <= 6 and 3 <= y <= 6: return True
        else:
            if 3 <= x <= 6 and 5 <= y <= 6: return True
    elif t == 2:
        if w == 1:
            if x == 4 and y == 4: return False
            if 4 <= x <= 6 and 4 <= y <= 6: return True
        elif w == 2:
            if x == 4 and y == 6: return False
            if 4 <= x <= 6 and 4 <= y <= 6: return True
        elif w == 3:
            if x == 6 and y == 6: return False
            if 4 <= x <= 6 and 4 <= y <= 6: return True
        else:
            if x == 6 and y == 4: return False
            if 4 <= x <= 6 and 4 <= y <= 6: return True

    return False



class colors:
    YELLOW = '\033[33m'
    BLUE = '\033[34m'
    RED = '\033[31m'
    CYAN = '\033[96m'
    GREEN = '\033[92m'
    ENDC = '\033[0m'


def bold_sym(s: str):
    return "╋" if s == "┼" else "┃" if s == "│" else "━━━" if s == "───" else s


def __str__(self):
    if self.st[0, 2] == 0:
        (px, py), (ox, oy) = self.position(), self.position(1)
    else:
        (px, py), (ox, oy) = self.position(1), self.position()

    res = ''
    board = [[("┼" if y % 2 else "│") if x % 2 else ("───" if y % 2 else "   ") for x in range(17)] for y in range(17)]

    board[2 * (py - 1)][2 * (px - 1)] = f'{colors.BLUE} O {colors.ENDC}'
    board[2 * (oy - 1)][2 * (ox - 1)] = f'{colors.RED} O {colors.ENDC}'

    def hl(x, y):
        if x >= 0 and x < 17 and y >= 0 and y < 17:
            board[y][x] = bold_sym(board[y][x])

    for wall_i in self.st[2:16]:
        wx, wy, opt = wall_i
        wx = 2 * wx - 1
        wy = 2 * wy - 1
        if opt == 1:
            for ty in [wy - 1, wy, wy + 1]:
                hl(wx - 2, ty)
        elif opt == 2:
            for tx in [wx - 1, wx, wx + 1]:
                hl(tx, wy - 2)

    for wall_l in self.st[16:]:
        wx, wy, opt = wall_l
        wx = 2 * wx - 1
        wy = 2 * wy - 1
        if opt == 1:
            for x, y in [(wx - 1, wy - 2), (wx - 2, wy - 2), (wx - 2, wy - 1)]: hl(x, y)
        elif opt == 2:
            for x, y in [(wx - 1, wy), (wx - 2, wy), (wx - 2, wy - 1)]: hl(x, y)
        elif opt == 3:
            for x, y in [(wx - 1, wy), (wx, wy), (wx, wy - 1)]: hl(x, y)
        elif opt == 4:
            for x, y in [(wx - 1, wy - 2), (wx, wy - 2), (wx, wy - 1)]: hl(x, y)

    for i, line in enumerate(board):
        res = f'  {"".join(line)}{colors.CYAN}{(i + 2) // 2 if not i % 2 else " "}{colors.ENDC}\n' + res

    res += f'   {"   ".join(map(lambda i: f"{colors.GREEN}{i}{colors.ENDC}", range(1, 10)))}'
    return res



from ...const import BOARD_SIZE, I_COUNT, L_COUNT


def vertical_block(self, x, y):
    for i in range(2, 2 + I_COUNT * 2):
        if self.st[i, 1] == y and (self.st[i, 0] == x or self.st[i, 0] == x - 1) and self.st[i, 2] == 2:
            return True

    for i in range(2 + I_COUNT * 2, 2 + (I_COUNT + L_COUNT) * 2):
        if self.st[i, 2] == 1 and self.st[i, 0] == x and self.st[i, 1] == y:
            return True
        if self.st[i, 2] == 2 and self.st[i, 0] == x and self.st[i, 1] == y - 1:
            return True
        if self.st[i, 2] == 3 and self.st[i, 0] == x and self.st[i, 1] == y - 1:
            return True
        if self.st[i, 2] == 4 and self.st[i, 0] == x and self.st[i, 1] == y:
            return True

    return False


def horizontal_block(self, x, y):
    for i in range(2, 2 + I_COUNT * 2):
        if self.st[i, 0] == x and (self.st[i, 1] == y or self.st[i, 1] == y - 1) and self.st[i, 2] == 1:
            return True

    for i in range(2 + I_COUNT * 2, 2 + (I_COUNT + L_COUNT) * 2):
        if self.st[i, 2] == 1 and self.st[i, 0] == x and self.st[i, 1] == y:
            return True
        if self.st[i, 2] == 2 and self.st[i, 0] == x and self.st[i, 1] == y:
            return True
        if self.st[i, 2] == 3 and self.st[i, 0] == x - 1 and self.st[i, 1] == y:
            return True
        if self.st[i, 2] == 4 and self.st[i, 0] == x - 1 and self.st[i, 1] == y:
            return True

    return False



import numpy as np
from ..const import BOARD_SIZE, I_COUNT, L_COUNT


class State:
    def __init__(self, state=None, dtype=np.uint8):
        if state is None:
            self.st = np.zeros((2 + (I_COUNT + L_COUNT) * 2, 3), dtype=dtype)
            self.st[0, 0] = (BOARD_SIZE + 1) // 2
            self.st[0, 1] = 1
            self.st[1, 0] = (BOARD_SIZE + 1) // 2
            self.st[1, 1] = BOARD_SIZE
        else:
            self.st = state

    def copy(self):
        return State(self.st.copy())

    def player(self, op=None):
        if op is None:
            return self.st[0, 2]
        return 1 - self.st[0, 2]

    def internal(self, op=None):
        pl = self.player()
        return 4 <= self.st[pl, 0] <= 6 and 4 <= self.st[pl, 1] <= 6

    def i_walls(self):
        return self.st[2 + np.array(*np.where(self.st[2:2 + I_COUNT * 2, 0] != 0)), :]

    def l_walls(self):
        return self.st[2 + I_COUNT * 2 + np.array(*np.where(self.st[2 + I_COUNT * 2:, 0] != 0)), :]

    def position(self, op=None):
        pl = self.player(op)
        return self.st[pl, 0], self.st[pl, 1]

    def win(self):
        if self.st[0, 1] == BOARD_SIZE:
            return 0
        elif self.st[1, 1] == 1:
            return 1
        return None

    from .validate.blocked import horizontal_block, vertical_block
    from .validate.movable import movable
    from .validate.endable import endable
    from .validate.cross import cross
    from .validate.internal import block_internal

    from .actions import move, place_i, place_l
    from .actions import act
    from .print import __str__



ACTION_MOVE = 0
ACTION_PLACE_I = 1
ACTION_PLACE_L = 2

from .move import move
from .place import place_i, place_l


def act(self, t, p1, p2, p3=None):
    if t == ACTION_MOVE:
        return self.move(p1, p2)
    if t == ACTION_PLACE_I:
        return self.place_i(p1, p2, p3)
    if t == ACTION_PLACE_L:
        return self.place_l(p1, p2, p3)
    return None


class ACTION:
    MOVE = ACTION_MOVE
    PLACE_I = ACTION_PLACE_I
    PLACE_L = ACTION_PLACE_L
